/*
 *  
 *  Copyright (c) 2004 Steve Slaven, All Rights Reserved.
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA 02111-1307 USA
 *  
*/
package net.hoopajoo.cheeselab3;
import java.util.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import net.hoopajoo.swingx.*;
import java.io.*;

// The main mapper panel
//class CheeseCanvas extends JPanel implements MouseListener,MouseMotionListener,KeyListener {
class CheeseCanvas extends JComponent implements MouseListener,MouseMotionListener,KeyListener {
	final static int TOOL_SELECT = 0;
	final static int TOOL_MERGE = 1;
	final static int TOOL_MOVE = 2;
	final static int TOOL_DELETE = 3;
	final static int TOOL_POINT = 4;
	final String [] toolnames = {
			"select_tool",
			"merge_tool",
			"move_tool",
			"delete_tool",
			"point_tool"
	};
	Cursor [] tool_cursors = {
			new Cursor( Cursor.DEFAULT_CURSOR ),
			new Cursor( Cursor.HAND_CURSOR ),
			new Cursor( Cursor.MOVE_CURSOR ),
			createCursor("delete"),
			new Cursor( Cursor.CROSSHAIR_CURSOR )
	};

	CheesePlayerData players[];
	CheeseMapData map;
	int selectedid = 0;
	int tool_active = TOOL_SELECT;
	String tool_saved = "tool_select";
	int ctx_edit_roomid = 0;
	Point3d drag_start;
	Point drag_start2d, drag_end2d;
	boolean lclick_dragging = false;
	boolean rclick_dragging = false;
	public PointToolCallback point_callback = null;
	Cursor cursor_saved = getCursor();
	Cursor cursor_movemap = new Cursor( Cursor.MOVE_CURSOR );
	Cursor cursor_busy = new Cursor( Cursor.WAIT_CURSOR );

	// Gridded overlays, TODO: make this a gui option
	boolean render_grid = true;
	
	// futures are buggy
	boolean render_future = false;
	
	// TODO: These have no ui yet
	boolean render_notes = false;
	int render_notes_length = 15;
	
	String timer_section;
	Date timer_start, timer_end;
	
	HashMap selected_rooms = new HashMap();
	
	int tile_width = 32,
		tile_height = 32,
		// Current origin info
		origin_x, origin_y, origin_z,

		// The origin coords at time of last map rendering
		// saved so that players and stuff overlay properly
		// when the map hasn't refreshed yet, otherwise they
		// will jump around the map during heavy movement
		map_origin_x, map_origin_y, map_origin_z,

		// The values returned by the server for the last map command,
		// these are initialized to zero.  If they are not zero then
		// the server supports sending the bounding box with the map
		// data and in turn we set it here so that you don't get
		// "partial" renders where the origin moved again since the
		// last map-request was completed causing your bounding box for the
		// map data vs the bounding box for the origin to be different
		// for example if you mash the navigator button fast, you will probably
		// move the origin more than the map response.  It is not fatal but
		// kind of ugly and this fixes the problem in a backwards compatible way
		served_origin_x = 0, served_origin_y = 0, served_origin_z = 0,

		rows = 0, cols = 0;
	String theme;
	private long changed = 0,
		// this basically tracks if we need to redraw the mapcache (new rooms, map data, resize, etc)
		changed_map = 0,
		last_map = 0,
		last_sync = 0;
	
	//private VolatileImage screencache = null;
	//private BufferedImage screencache = null;
	private BufferedImage screencache = null;
	private BufferedImage mapcache = null;

	CheeseRenderer renderer;

	// TODO: Needs another map for temp overlays, that can be image/text
	// so that we can "blink" people off where.  These overlays
	// should have an expire time (+5 seconds?) that will make them
	// auto-delete at the next render after the expiry time
	// this would also make the "hud" display of hp/move/mana possible

	// These are by x,y,z key
	HashMap mapdata = new HashMap();
	HashMap mtimedata = new HashMap();
	
	// This contains the roomid -> x,y,z key so that 
	// when new room data comes in, it can be deleted from the cache
	HashMap reversedata = new HashMap();

	// This is here so that the tilenames are standard for all renderers
	HashMap tilenames = new HashMap();

	// HashMap has faster get but slower put apparently, but it'll work
	private HashMap image_cache = new HashMap();

	CheeseCanvas() {
		addMouseListener( this );
		addMouseMotionListener( this );
		addKeyListener( this );
		setToolTipText( "Room Info" );

		// Set tile mapping, this provides common tile names for all renderers
		tilenames.put("@","indoors");
		tilenames.put("<","mountain");
		tilenames.put("+","road");
		tilenames.put("#","city");
		tilenames.put("~","shallow_water");
		tilenames.put("\"","brush");
		tilenames.put("_","field");
		tilenames.put("w","water");
		tilenames.put("=","tunnel");
		tilenames.put("%","forest");
		tilenames.put("W","rapids");
		tilenames.put("O","cavern");
		tilenames.put("(","hills");
		tilenames.put("U","underwater");

		String r = CheeseMap.prop_get_string( "renderer" );
		if( r.equals( "cached" ) ) {
			CheeseLab.dmesg( 1, "Enabling cached renderer" );
			renderer = new CheeseRendererCached();
		}else if( r.equals( "basic" ) ) {
			CheeseLab.dmesg( 1, "Enabling basic renderer" );
			renderer = new CheeseRendererBasic();
		}else{
			// Default for wacky values
			CheeseLab.dmesg( 1, "Unknown renderer: '" + r + "', using base, sucks for you" );
			renderer = new CheeseRenderer();
		}
		renderer.set_canvas( this );

		setDoubleBuffered( false );
		setFocusable( true );
		// requestDefaultFocus();
		// requestFocus();
		requestFocusInWindow();
	}

	public void update_map( byte[] buffer ) {
		synchronized ( mapdata ) {
			String s = new String( buffer );
			BufferedReader in = new BufferedReader( new StringReader( s ) );
			String line;
			try{
				while( (line = in.readLine()) != null ) {
					String[] values = line.split( "," );

					// If values.length > 5, then this line is actually the
					// map boundry line (e.g. 0,0,0,0,0,x,y,z,xx,yy,zz)
					if( values.length > 5 ) {
						served_origin_x = Integer.parseInt( values[ 5 ] );
						served_origin_y = Integer.parseInt( values[ 6 ] );
						served_origin_z = Integer.parseInt( values[ 7 ] );
					}else{
						String roomkey = values[ 1 ] + "," + values[ 2 ] + "," + values[ 3 ];
						Integer roomid = new Integer( values[ 0 ] );

						// Check if there is old data to clear
						String oldkey = (String)reversedata.get( roomid );
						if( oldkey != null ) {
							mapdata.remove( oldkey );
							mtimedata.remove( oldkey );
						}

						// Real map data
						mapdata.put( roomkey, roomid );
						mtimedata.put( roomkey, new Long( values[ 4 ] ) );
					}
				}
			}catch( Exception e ){
				e.printStackTrace();
			}
		}

		map_changed();
		changed( true );
	}
	
	public void mouseMoved(MouseEvent e) {
	}
	
	public void mouseDragged(MouseEvent e) {
		if( lclick_dragging && tool_active == TOOL_SELECT ) {
			// draggin to select
			drag_end2d = e.getPoint();
			repaint();
		}
		if( rclick_dragging ) {
			// dragging the map around
			drag_end2d = e.getPoint();
			repaint();
		}
	}
	
	public void mousePressed(MouseEvent e) {
		java.awt.Point p = e.getPoint();
		
		if( e.getButton() == 1 ) {
			CheeseLab.dmesg( 3, "Button 1 down" );

			drag_start = local_to_global(p.x, p.y);
			drag_start2d = p;
			drag_end2d = p; // ends here too atm
			lclick_dragging = true;
			
		}else if( e.getButton() == 3 ) {
			CheeseLab.dmesg( 3, "Button 3 down" );
			drag_start = local_to_global( p.x, p.y );
			drag_start2d = p;
			drag_end2d = p; // ends here too atm
			rclick_dragging = true;
			cursor_saved = getCursor();
			setCursor( cursor_movemap );
		}
	}

	public void mouseReleased(MouseEvent e) {
		java.awt.Point p = e.getPoint();			
		Point3d drag_end = local_to_global( p.x, p.y );

		if( e.getButton() == 1 ) {
			if( tool_active == TOOL_SELECT ) {
				boolean select_3d = false;
				
				if(!e.isShiftDown()) 	// Allow adding rooms to selection by holding the SHIFT key
					selected_rooms.clear();
				
				if(e.isControlDown())	// Do 3D select if Control was held down
					select_3d = true;
				
				CheeseLab.dmesg( 3, "Button 1 up" );
				
				int x, y, z,
					min_x, max_x,
					min_y, max_y,
					min_z, max_z;
				
				min_z = max_z = drag_start.z;
				
				if( drag_start.x < drag_end.x ) {
					min_x = drag_start.x;
					max_x = drag_end.x;
				}else{
					min_x = drag_end.x;
					max_x = drag_start.x;
				}
				
				if( drag_start.y < drag_end.y ) {
					min_y = drag_start.y;
					max_y = drag_end.y;
				}else{
					min_y = drag_end.y;
					max_y = drag_start.y;
				}
				
				/* TODO: Actually implement 3D select. We need to make selection algorithm more
				 *       efficent. One way is to implement search by coordinates in broker.
				 */
				for( z = min_z; z <= max_z; z++ ) {
					CheeseLab.dmesg( 3, "Selecting rooms from: " + min_x + "," + min_y + "," + z +
							" " + max_x + "," + max_y + "," + z );
					for( x = min_x; x <= max_x; x++ ) {
						for( y = min_y; y <= max_y; y++ ) {
							// These rooms ought to be in our map since you can't really
							// click on them unless you can see them
							String global_key = new String( x + "," + y + "," + z );
							Integer roomid = (Integer)mapdata.get( global_key );
							if( roomid != null ) {
								CheeseLab.dmesg( 3, "Selecting: " + roomid );
								selected_rooms.put( roomid, roomid );
							}
						}
					}
				}
				
				/* TODO: Maybe we should move map if there was only one room selected.
				 *       The need for LMB navigation is debatable though
				 * 
				 * if( selected_rooms.size() == 1 ) {
				 *	Point3d p3d = local_to_global( p.x, p.y );
				 *	center_on( p3d.x, p3d.y, p3d.z );
				 *	CheeseMap.set_follow_player( false );
				 * }
				 */
				
				requestFocus();
			}else if( tool_active == TOOL_DELETE ) {
				// delete
				String global_key = new String( drag_end.x + "," + drag_end.y + "," + drag_end.z );
				Integer roomid = (Integer)mapdata.get( global_key );
				if( roomid != null )
					CheeseMap.send_tx( 1, "delete", roomid.toString() );
			}else if( tool_active == TOOL_MERGE ) {
				// merge
				String global_key = new String( drag_end.x + "," + drag_end.y + "," + drag_end.z );
				Integer roomid = (Integer)mapdata.get( global_key );
				if( roomid != null )
					CheeseMap.send_tx( 1, "merge", roomid.toString() );
			}else if( tool_active == TOOL_POINT ) {
				// Restore last tool. Execute callback.
				setEditTool( tool_saved );
				if (point_callback != null) {
					point_callback.pointedTo(p);
					point_callback = null;
				}
			}
			
			lclick_dragging = false;
			map_changed();
			changed();
		}
		
		if( e.getButton() == 3 ) {
			int dx = drag_start.x - drag_end.x;
			int dy = drag_start.y - drag_end.y;

			rclick_dragging = false;
			setCursor( cursor_saved );

			if( dx != 0 || dy != 0 ) {
				move_origin( dx, dy, 0 );
				CheeseMap.set_follow_player( false );
			}else{
				// Call repaint to cancel visual effect of "accidental" dragging
				repaint();
				
				// Now build and open context menu
				JPopupMenu menu = new JPopupMenu();
				JMenuItem item;
				CheeseRoomData room = null;

				// Context menu room editor
				String r_name;
				Integer roomid = (Integer)mapdata.get(drag_end.x + ","
						+ drag_end.y + "," + drag_end.z);
				if( roomid != null ) {
					room = map.get_room( roomid.intValue() );
					r_name = room.name;
					ctx_edit_roomid = roomid.intValue();
				}else{
					r_name = "No Room Here";
					ctx_edit_roomid = 0;
				}

				// Trim room name length to show it in context menu
				if( r_name.length() > 25 )
					r_name = r_name.substring( 0, 22 ) + "... ";

				item = new JMenuItem( r_name );
				item.setEnabled( false );
				menu.add( item );
				menu.addSeparator();

				if( ctx_edit_roomid != 0 ) {
					item = new JMenuItem( "Open Room Editor" );
					item.addActionListener( new ActionListener() {
						public void actionPerformed(ActionEvent e) {
							selectedid = ctx_edit_roomid;
							CheeseLab.dmesg(3, "Do editor for: " + ctx_edit_roomid);
							CheeseMap.set_edit_room(map.get_room(ctx_edit_roomid));
						}
					} );
					menu.add(item);
					
					// Construct sub-menu for Overlay editing
					String [] ovrl = CheeseMapRoomEditor.common_overlays;
					JMenu submenu = new JMenu( "Overlays" );
					for( int i = 0; i < ovrl.length; i++) {
					    boolean checked = false; 
					    ImageIcon icon = new ImageIcon(
					            getImage( "overlay_" + ovrl[i] ).
					            	getScaledInstance( 16, 16,
					            	        Image.SCALE_SMOOTH ) );
					    for( int j = 0; j < room.overlays.length; j++ ) {
					        if( room.overlays[j].equals( ovrl[i] ) )
					            checked = true;
					    }
					    item = new JCheckBoxMenuItem( ovrl[i], icon, checked );
					    item.addActionListener( new CtxEditorListener( 
					            		ovrl[i], ctx_edit_roomid ) );
					    item.setActionCommand( "set_overlay" );
					    submenu.add( item );
					}
					menu.add( submenu );

					// Construct sub-menus for exit editing
					String[] dirs = { "N", "S", "E", "W", "U", "D" };
					int[] cur_exits = { room.exit_n, room.exit_s, room.exit_e,
							room.exit_w, room.exit_u, room.exit_d };

					for( int i = 0; i < dirs.length; i++ ) {
						JMenu m = new JMenu( "Exit " + dirs[i] );
						ButtonGroup bg = new ButtonGroup();

						bg.add( item = new JRadioButtonMenuItem( "None" ) );
						if( cur_exits[i] == 0 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_none" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "Unexplored" ) );
						if( cur_exits[i] == -1 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_unexplored" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "Death Trap" ) );
						if( cur_exits[i] == -2 )
							item.setSelected( true );
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_dt" );
						m.add( item );
						bg.add( item = new JRadioButtonMenuItem( "To room ..." ) );
						if( cur_exits[i] > 0 ) {
							item.setText( "To room " + cur_exits[i] );
							item.setSelected( true );
						}
						item.addActionListener( new CtxEditorListener( dirs[i], ctx_edit_roomid ) );
						item.setActionCommand( "set_exit_from_map" );
						m.add( item );

						menu.add( m );
					}
				}

				// Main operations
				menu.addSeparator();
				menu.add( new JSelectedCheckMenuItem( CheeseMap.follow_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.levelup_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.leveldown_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.zoomin_action ) );
				menu.add( new JSelectedMenuItem( CheeseMap.zoomout_action ) );

				item = new JMenuItem( "Merge Selected" );
				item.addActionListener( new ActionListener() {
					public void actionPerformed(ActionEvent e) {
						CheeseMap.canvas.merge_selected();
					}
				} );
				menu.add(item);
				
				// Toolbox
				menu.add( CheeseMap.toolbox.createMenu( "Tools" ) );

				// Phew! Done! Show it.
				// menu.setLightWeightPopupEnabled(false);
				menu.show( this, p.x, p.y );
			}
		}
	}
	
	public void mouseEntered(MouseEvent e) { }
	public void mouseExited(MouseEvent e) { }
	public void mouseClicked(MouseEvent e) { }

	public String getToolTipText( MouseEvent e ) {
		java.awt.Point p = e.getPoint();

		Point3d p3d = local_to_global( p.x, p.y );
		
		// TODO: make this useful :p
		String ret = new String();
		Integer roomid = (Integer)mapdata.get( p3d.x + "," + p3d.y + "," + p3d.z );
		Long mtime = (Long)mtimedata.get( p3d.x + "," + p3d.y + "," + p3d.z );
		if( roomid != null ) {
			CheeseRoomData room = map.get_room( roomid.intValue(), mtime.longValue() );
			String show_notes = room.notes;
			if( show_notes == null )
				show_notes = "";
	
			if( ! show_notes.equals( "" ) )
				show_notes = "<br>" + show_notes;

			// Check for hidden exits
			String show_hidden = "";
			if( ( ! room.door_n.equals( "" ) ) && ( ! room.door_n.equals( "exit" ) ) )
				show_hidden += "<br>Door north: " + room.door_n;

			if( ( ! room.door_s.equals( "" ) ) && ( ! room.door_s.equals( "exit" ) ) )
				show_hidden += "<br>Door south: " + room.door_s;

			if( ( ! room.door_e.equals( "" ) ) && ( ! room.door_e.equals( "exit" ) ) )
				show_hidden += "<br>Door east: " + room.door_e;

			if( ( ! room.door_w.equals( "" ) ) && ( ! room.door_w.equals( "exit" ) ) )
				show_hidden += "<br>Door west: " + room.door_w;

			if( ( ! room.door_u.equals( "" ) ) && ( ! room.door_u.equals( "exit" ) ) )
				show_hidden += "<br>Door up: " + room.door_u;

			if( ( ! room.door_d.equals( "" ) ) && ( ! room.door_d.equals( "exit" ) ) )
				show_hidden += "<br>Door down: " + room.door_d;

			ret = "<html>" + 
				"#" + room.roomid + " - " + room.x + "," + room.y + "," + room.z + "<br>" + 
				room.name + 
				show_notes +
				show_hidden +
				"</html>";
		}else{
			//ret = "No Room Here";
			ret = null;
		}

		return( ret );
	}

	public void paintComponent(Graphics g) {
		if( CheeseMap.online == false ) {
			super.paintComponent( g );
			return;
		}
		
		// Handle right-click map move mode
		if( rclick_dragging ) {
			if( screencache != null ) // We can't move around nonexisting picture right?
				g.drawImage( screencache, drag_end2d.x - drag_start2d.x,
						drag_end2d.y - drag_start2d.y, null );
			return;
		}
		
		// Handle resize refreshes
		// TODO: take player position into account if in follow_player mode
		if( screencache != null ) {
			// validate width/height
			if( screencache.getWidth(null) != getWidth() ) {
				screencache.flush();
				screencache = null;
			}else if( screencache.getHeight(null) != getHeight() ) {
				screencache.flush();
				screencache = null;
			}
		}

		// If no screencache, then we are definately gonna need to repaint
		//if( screencache == null )
		//	changed( false );
		
		// Paint to a bitmap, that way exposed paint events don't flicker like mad
		if( changed != last_sync || screencache == null ) {
			CheeseLab.dmesg( 2, "Something changed: c" +
				changed + "/" + last_sync + " r" +
				changed_map + "/" + last_map  );

			last_sync = changed;

			// Freeze players
			players = map.get_players();

			// Make our double buffer
			if( screencache == null ) {
				// make cache object
				//screencache = createImage( getWidth(), getHeight() );
				CheeseLab.dmesg( 2, "Creating new buffered image for screen cache" );
				screencache = (BufferedImage)createImage( getWidth(), getHeight() );
				remap();
			}

			// Make sure the map matches the size of it
			if( mapcache != null ) {
				if( mapcache.getWidth() != screencache.getWidth() ) {
					mapcache.flush();
					mapcache = null;
				}else if( mapcache.getHeight() != screencache.getHeight() ) {
					mapcache.flush();
					mapcache = null;
				}
			}

			// Create the mapcache if needed
			if( mapcache == null ) {
				mapcache = (BufferedImage)createImage( screencache.getWidth(),
						screencache.getHeight() );
				map_changed();
			}

			// Initialize this render frame
			renderer.prepare();

			// Update base map if we got new map data
			if( last_map < changed_map ) {
				/*
				if( served_origin_x > 0 && served_origin_y > 0 && served_origin_z > 0 ) {
					map_origin_x = served_origin_x;
					map_origin_y = served_origin_y;
					map_origin_z = served_origin_z;
				}else{
					map_origin_x = origin_x;
					map_origin_y = origin_y;
					map_origin_z = origin_z;
				}
				*/
				map_origin_x = origin_x;
				map_origin_y = origin_y;
				map_origin_z = origin_z;

				last_map = changed_map;

				Date rstart = new Date();

				renderer.draw_map( mapcache );

				Date rend = new Date();
				long rdiff = rend.getTime() - rstart.getTime();
				double secs = (double)rdiff / 1000;
				CheeseLab.dmesg( 1, "Render time: " + rdiff + "ms (" + secs + " secs)" );
			}

			// Now render map data + players + hud to screencache
			Graphics2D sg = (Graphics2D)screencache.getGraphics();
			sg.drawImage( mapcache, 0, 0, null );
			sg.dispose();

			renderer.draw_players( screencache );
			renderer.draw_hud( screencache );
		}

		g.drawImage( screencache, 0, 0, null );
		
		// if dragging draw that
		if( lclick_dragging ) {
			CheeseLab.dmesg( 3, "Drawing rect" );
			g.setColor( Color.RED );
			g.drawRect( drag_start2d.x, drag_start2d.y,
					drag_end2d.x - drag_start2d.x,
					drag_end2d.y - drag_start2d.y );
		}
	}

	Image getImage( String name ) {
		return( getImage( name, false ) );
	}
	
	Image getImage( String name, boolean gridded ) {
		//System.out.println( "Getting tile: " + name );

		// Load an image
		// 1) Request and load from server first, then save to file
		// 2) Load from disk, search in local/theme, local, theme/

		// TODO: request from server? or no server-side tiles?

		// If not in cache load it
		//Image im = (Image)image_cache.get( name );
		Image im = (Image)image_cache.get( name );
		//System.out.println( "Got ptr: " + im );
		if( im == null ) {
			CheeseLab.dmesg( 3, "Image not in cache, getting" );
			Toolkit toolkit = Toolkit.getDefaultToolkit();
			
			// TODO: Try file in tiles/name.png then cacheprefix/tiles/name.png
			String tilenames[] = { 
				CheeseMap.cachedir + "/tiles/" + name + ".png",
				"tiles/" + theme + "/" + name + ".png",
				"tiles/default/" + name + ".png",
				"tiles/" + name + ".png"
			};

			CheeseLab.dmesg( 3, "Loading: " + name );
			for( int i = 0; i < tilenames.length; i++ ) {
				ImageIcon ii = new ImageIcon( tilenames[ i ] );
				if( ii.getIconWidth() != -1 ) {
					CheeseLab.dmesg( 3, "Got match: " + tilenames[ i ] );
					im = ii.getImage().getScaledInstance( tile_width, tile_height, Image.SCALE_SMOOTH );
					
					// Force loading
					while( ! prepareImage( im, -1, -1, this ) );
			
					// Now convert to a compatible image from this gc
					Graphics2D sg = screencache.createGraphics();
					
					// Alpha blending does not do acceleration, but I like it, and under non-win32 environments the accel doesn't work anyways :(
					// so you all have to suffer!! *hop*
					Image t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.BITMASK );
					//Image t = sg.getDeviceConfiguration().createCompatibleImage( tile_width, tile_height, Transparency.TRANSLUCENT );
					Graphics gg = t.getGraphics();
					gg.drawImage( im, 0, 0, null );
					
					if( gridded )
						gg.drawImage( getImage( "grid" ), 0, 0, null );
						
					gg.dispose();
					sg.dispose();

					im = t;
				}
			}
			
			// Cache it in memory if we got one
			if( im != null ) {
				image_cache.put( name, im );
			}
		}

		// if we still don't have an image, return missing
		if( im == null )
			return( getImage( "missing" ) );

		return( im );
	}
	
	public Point global_to_local_reference( int x, int y, int z, int rx, int ry, int rxz ) {
		Point p = new Point();
		p.x = ( x - rx ) * tile_width;
		//p.y = ( rows - ( y - origin_y ) ) * tile_height;
		p.y = ( y - ry ) * tile_height;
		return( p );
	}
	
	public Point global_to_local( int x, int y, int z ) {
		return( global_to_local_reference( x, y, z, origin_x, origin_y, origin_z ) );
	}
	
	public Point3d local_to_global( int x, int y ) {
		Point3d p = new Point3d();
		
		p.x = origin_x + ( x / tile_width );
		//p.y = origin_y + ( rows - ( y / tile_width ) );
		p.y = origin_y + ( y / tile_height );
		p.z = origin_z;
		
		return( p );
	}
	
	public void set_map( CheeseMapData m ) {
		map = m;
	}

	public void center_on( int x, int y, int z ) {
		CheeseLab.dmesg( 2, "Centering on: " + x + "," + y + "," + z );
		set_origin( x - ( cols / 2 ), y - ( rows / 2 ), z );
	}
	
	public void move_origin( int dx, int dy, int dz ) {
		CheeseLab.dmesg( 2, "Moving: " + dx + "," + dy + "," + dz );
		set_origin( origin_x + dx, origin_y + dy, origin_z + dz );
	}
	
	public void set_origin( int x, int y, int z ) {
		CheeseLab.dmesg( 2, "Setting origin: " + x + "," + y + "," + z );
		origin_x = x;
		origin_y = y;
		origin_z = z;

		remap();
		changed();
	}

	// Request updated map data, this should be called whenever the
	// origin moves or if rooms are moved around (world change events)
	// This will in-advertantly trigger a repaint
	public void remap() {
		// We mark the sync now so that if there are changed while we
		// are rendering, we will get them
		int rend_x = origin_x,
			rend_y = origin_y,
			rend_z = origin_z;

		CheeseLab.dmesg( 2, "Remap" );

		cols = getWidth() / tile_width;
		rows = getHeight() / tile_height;

		CheeseLab.dmesg( 3, "Size: " + rows + "," + cols );
		CheeseLab.dmesg( 3, "Origin: " + rend_x + "," + rend_y + "," + rend_z );
			
		// Figure our bounding box
		int start_x = rend_x;
		int start_y = rend_y;
		int start_z = rend_z;

		int end_x = start_x + cols;
		int end_y = start_y + rows;
		int end_z = rend_z + 1;

		// Request an update from the server, when it's receieved another repaint will be issued
		CheeseMap.send_tx( 1, "map", 
			start_x + "," + start_y + "," + start_z + " " +
			end_x + "," + end_y + "," + end_z );
	}

	public void map_changed() {
		changed_map++;
		//CheeseLab.dmesg( 1, "room_changed()" );
	}

	public void changed() {
		changed( true );
	}
	
	public void changed( boolean redraw ) {
		changed++;

		//CheeseLab.dmesg( 1, "changed()" );

		if( redraw )
			repaint( 100 );
	}

	public void keyTyped( KeyEvent e ) {	}
	public void keyReleased( KeyEvent e ) { }

	public void keyPressed( KeyEvent e ) {
		// If we have selected items
		if( ! selected_rooms.isEmpty() ) {
			int x = 0,
				y = 0,
				z = 0;
			int delta = 1;
			
			if( (e.getModifiers() & InputEvent.SHIFT_MASK) != 0 ) {
				delta = 5;
			}
			
			if( e.getKeyCode() == KeyEvent.VK_UP ) {
				y -= delta;
			}else if( e.getKeyCode() == KeyEvent.VK_DOWN ) {
				y += delta;
			}else if( e.getKeyCode() == KeyEvent.VK_LEFT ) {
				x -= delta;
			}else if( e.getKeyCode() == KeyEvent.VK_RIGHT ) {
				x += delta;
			}else if( e.getKeyCode() == KeyEvent.VK_PAGE_UP ) {
				z++;
			}else if( e.getKeyCode() == KeyEvent.VK_PAGE_DOWN ) {
				z--;
			}
			
			if( x != 0 || y != 0 || z != 0 ) {
				CheeseLab.dmesg( 3, "Gonna move rooms: " + x + "," + y + "," + z );
				
				Object[] roomids = selected_rooms.keySet().toArray();
				
				// Loop over selected rooms and send save commands
				String tx = x + "," + y + "," + z;
				for( int i = 0; i < roomids.length; i++ ) {
					int roomid = ((Integer)roomids[ i ]).intValue();
					CheeseLab.dmesg( 4, "  > " + roomid );
					tx += "," + roomid;
				}
				CheeseMap.send_tx( 1, "move", tx );
				changed();
				// TODO: make sure if we are EDITING one of these that we update the room editor!
				// it might be easeier if the room tx handler checked to see if the room is in the
				// editor and just passed it on magically
			}
		}
	}
	
	public void set_tile_size( int width, int height ) {
		tile_width = width;
		tile_height = height;
		image_cache.clear();
		map_changed();
		changed();
	}
	
	private void _start_timer( String section ) {
		timer_start = new Date();
		timer_section = section;
		System.err.println( "Timer started: " + section );
	}
	
	private void _end_timer() {
		timer_end = new Date();
		long diff = timer_end.getTime() - timer_start.getTime();
		double secs = (double)diff / 1000;
		System.err.println( timer_section + " took: " + diff + "ms (" + secs + " secs)" );
	}

	// Perform merge for all items on screen
	public void merge_all() {
		cursor_saved = getCursor();
		setCursor( cursor_busy );

		Object[] roomids = mapdata.values().toArray();
		for( int i = 0; i < roomids.length; i++ )
			CheeseMap.send_tx( 1, "merge", ((Integer)roomids[ i ]).toString() );

		setCursor( cursor_saved );
	}

	// Perform merge on all selected items
	public void merge_selected() {
		cursor_saved = getCursor();
		setCursor( cursor_busy );

		Object[] roomids = selected_rooms.keySet().toArray();
		for( int i = 0; i < roomids.length; i++ )
			CheeseMap.send_tx( 1, "merge", ((Integer)roomids[ i ]).toString() );

		setCursor( cursor_saved );
	}

	public void setEditTool(String toolname) {
		if ( toolname.equals( toolnames[TOOL_SELECT] ) ) {
			setCursor( tool_cursors[TOOL_SELECT] );
			tool_active = TOOL_SELECT;
		}else if ( toolname.equals( toolnames[TOOL_MERGE] ) ) {
			setCursor( tool_cursors[TOOL_MERGE] );
			tool_active = TOOL_MERGE;
		}else if ( toolname.equals( toolnames[TOOL_MOVE] ) ) {
			setCursor( tool_cursors[TOOL_MOVE] );
			tool_active = TOOL_MOVE;
		}else if ( toolname.equals( toolnames[TOOL_DELETE] ) ) {
			setCursor( tool_cursors[TOOL_DELETE] );
			tool_active = TOOL_DELETE;
		}else if ( toolname.equals( toolnames[TOOL_POINT] ) ) {
			setCursor( tool_cursors[TOOL_POINT] );
			tool_saved = getEditTool();
			tool_active = TOOL_POINT;
		}
	}
	
	public String getEditTool() {
		return toolnames[tool_active];
	}
	
	private void save_exit(int from, String dir, int to) {
		CheeseLab.dmesg( 3, "Save exit: " + from + " - " + dir + " -> " + to );

		// FIXME: This has an unsolvable race condition.
		//        basically what we do is we fetch the room info and then send it back.
		//        This is a waste of bandwidth and falls apart if some1 changes
		//        room data in the meantime. Also waiting for a ping freezes the GUI.
		//        Creating an atomic exit change command in broker would be a lot
		//        better idea
		//        eg. CheeseMap.send_tx( 1, "save_exit", from + "," + dir + "," + to );

		// Make sure it gets in the cache
		CheeseRoomData room = map.get_room( from, System.currentTimeMillis() );
		CheeseMap.send_tx(1, "ping", "set_edit_room");
		while ( !CheeseMap.checkping() );

		room.mtime = System.currentTimeMillis();
		if( dir.equalsIgnoreCase( "u" ) )
			room.exit_u = to;
		else if( dir.equalsIgnoreCase( "d" ) )
			room.exit_d = to;
		else if( dir.equalsIgnoreCase( "n" ) )
			room.exit_n = to;
		else if( dir.equalsIgnoreCase( "s" ) )
			room.exit_s = to;
		else if( dir.equalsIgnoreCase( "e" ) )
			room.exit_e = to;
		else if( dir.equalsIgnoreCase( "w" ) )
			room.exit_w = to;

		// Send save room command
		CheeseMap.send_tx( 1, "save", CheeseDecoder.encode_object(room) );

		changed();
	}
	
	private void save_overlay(int from, String ovrl, boolean en) {
	    Vector novrls = new Vector();
	    
	    CheeseLab.dmesg( 3, "Save overlay: " + from + " - " + ovrl + " -> " + en );
	    
		// Make sure it gets in the cache
		CheeseRoomData room = map.get_room( from, System.currentTimeMillis() );
		CheeseMap.send_tx(1, "ping", "set_edit_room");
		while ( !CheeseMap.checkping() );

		room.mtime = System.currentTimeMillis();
		// Copy all overlays except current one to vector 
		for( int i = 0; i < room.overlays.length; i++ ) {
		    String o = room.overlays[i];
		    if( !o.equals("") && !o.equals( ovrl ) )
		        novrls.add( o );
		}		
		// Add current overlay if requested so
		if( en )
		    novrls.add( ovrl );
		room.overlays = (String [])novrls.toArray( new String[0] ); 
		
		// Send save room command
		CheeseMap.send_tx( 1, "save", CheeseDecoder.encode_object(room) );

		changed();
	}
	
	/**
	 * Create custom cursor from a graphics file
	 * @param name  name to be looked up. The filename takes form: ui/ui_cursor_name.png
	 * @return      Cursor object with new cursor.
	 */
	Cursor createCursor(String name) {
	    // TODO: Scale cursor to system preferred size
		Toolkit tk = Toolkit.getDefaultToolkit();
		Image img = new ImageIcon( "ui/ui_cursor_" + name + ".png" ).getImage();
		return tk.createCustomCursor(img, new Point( 0,0 ), name);
	}
	
	/**
	 * Implement this class to add callback to point tool. This tool can be used
	 * to allow user to select just one room (to point a room).
	 */
	abstract public class PointToolCallback {
		abstract void pointedTo(Point p);
	}

	/**
	 * This class is used to implement Context Menu editor commands.
	 */
	class CtxEditorListener implements ActionListener {
		String sdata;

		int roomid;

		CtxEditorListener(String sdata, int roomid) {
			this.sdata = sdata;
			this.roomid = roomid;
		}

		public void actionPerformed(ActionEvent e) {
			String cmd = e.getActionCommand();
			if(cmd.equals( "set_exit_none" )) {
				save_exit( roomid, sdata, 0 );
			}else if( cmd.equals( "set_exit_unexplored" ) ) {
				save_exit( roomid, sdata, -1 );
			}else if( cmd.equals( "set_exit_dt" ) ) {
				save_exit( roomid, sdata, -2 );
			}else if( cmd.equals( "set_exit_from_map" ) ) {
				point_callback = new PointToolCallback() {
					void pointedTo(Point p) {
						Point3d p3d = local_to_global( p.x, p.y );

						Integer room = (Integer)mapdata.get( p3d.x + ","
								+ p3d.y + "," + p3d.z );
						if( room != null ) {
							save_exit( roomid, sdata, room.intValue() );
						}
					}
				};
				setEditTool( "point_tool" );
			}else if( cmd.equals( "set_overlay" )) {
			    JCheckBoxMenuItem src = (JCheckBoxMenuItem)e.getSource();
			    save_overlay( roomid, sdata, src.isSelected() );
			}
		}
	}

	public String tilename( CheeseRoomData r ) {
		String tilename = (String)tilenames.get( r.terrain );
		if( tilename == null )
			tilename = "unknown";
		return( tilename );
	}
}
